# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Plugin module which does an ML inference from a Google ML model via REST for each file offered on its asset """

import os
import subprocess
import requests
import json

import logging
from foglamp.common import logger
import filter_ingest

from foglamp.common.plugin_helpers import HandleMap, PluginHandle

import filter_ingest

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

FOGLAMP_ROOT=os.getenv("FOGLAMP_ROOT", "/usr/local/foglamp")
FOGLAMP_DATA=os.getenv("FOGLAMP_DATA", os.path.join(FOGLAMP_ROOT, "data"))

_LOGGER = logger.setup(__name__, level=logging.INFO)
_LOGGER.info("DONE LOAD PLUGIN")

_handle_map = HandleMap(__name__)


class EdgeMLHandle(PluginHandle):
    """
    Stores state passed to/from Foglamp callbacks to manage working with inference engine
    """

    def __init__(self, config, ingest_ref, callback):
        super().__init__("edge-ml", (ingest_ref, callback))

        if True:
            self.logger.setLevel(logging.DEBUG)
            
        # convert config entries into handle attributes
        self.config_update(config)

    def config_update(self, udict):
        """ config_update -- configure the handle member values as indicated by the update dict """

        # do generic updates
        super().config_update(udict)

        # fix-up local specializations
        if self.out_asset_name == "":
            self.out_asset_name = self.asset_name

        if self.edge_ml_url == "":
            self.edge_ml_url = self._get_edgeml_url()
    
    EDGEML_REST_RESOURCE="v1/vision/detection"
    DEPLOYMENT_NAME = os.getenv("DEPLOYMENT_NAME", "")

    def _get_edgeml_url(self):
        """ _get_edgeml_url - pull the REST url out of the kubernetes deployment information """

        _deployment_name = self.deployment_name if (self.deployment_name != '') else EdgeMLHandle.DEPLOYMENT_NAME
        self.logger.info("DEPLOY NAME  self:{} EdgeMLHandle: {}".format(self.deployment_name, EdgeMLHandle.DEPLOYMENT_NAME))
        if _deployment_name == "":
            raise ValueError("unspecified edgeMlUrl and no DEPLOYMENT_NAME environment or config setting")
        
        _p = subprocess.run(["kubectl", "get", "svc", _deployment_name,
                             "-o", "jsonpath='{.spec.clusterIP}:{.spec.ports[0].port}'"],
                            stdout=subprocess.PIPE,
                            check=True)
        _url = "http://{}/{}".format(eval(_p.stdout), EdgeMLHandle.EDGEML_REST_RESOURCE)
        return _url

    def get_inferences(self, data) -> [dict]:
        """ for each data element, for all readings, perform inference on the 'dir'/'name' images """

        self.logger.info("get_inferences: {}".format(data))

        def _infer(path) -> dict:
            """ _infer -- actually post the image data to the ml model, get the inference back """
            _inference = None
            try:
                self.logger.debug("{} -> {}".format(path, self.edge_ml_url))
                with open(path, 'rb') as fp:
                    _files = {'image': fp}
                    
                    r = requests.post(self.edge_ml_url, files=_files)

                    if r.status_code == 200:
                        _inference = r.json()
                        self.logger.debug("Inference returns: {}".format(_inference))
                        
                        _inference.update({'path': path})

                        if self.rm_file:
                            try:
                                os.unlink(path)
                            except Exception as ex:
                                self.logger.error("couldn't unlink {} ({})".format(path, ex))
                    else:
                        self.logger.error("POST {} returned {}/{}".format(_path, r.status_code, r.reason))
            except FileNotFoundError:
                self.logger.error("edgeml file not found {}".format(_path))
            except Exception as ex:
                self.logger.exception("edgeml exception {}".format(ex))
                raise
            
            return _inference

        ### main logic -- for each dir/name, create output from the inference from google ml ###

        # ASSUME data is a list of readings
            
        _inferences = []
        for d in data:
            # readings are {'asset': <in-asset>, 'readings': {'dir': dir, 'name': <fname>}...}
            if d['asset'] != self.asset_name:
                continue

            _r = d['readings']
            if 'name' not in _r or 'dir' not in _r:
                self.logger.error("'name' or 'dir' not found in readings keys {}".format(list(_r.keys())))
                continue

            _dir = _r['dir'].strip('"')
            _name = _r['name'].strip('"')
            _path = os.path.join(FOGLAMP_DATA, _dir, _name)

            _inference = _infer(_path)
            if _inference:
                _inference['img_name'] = _name
                _inferences.append(_inference)

        return _inferences

    def ingest(self, data):
        self.logger.debug("ingest: enable {}".format(self.enable))

        # data starts out with one or more readings
        if type(data) is dict:
            data = [data]
        
        if self.enable:

            # perform inference on stream of input file names
            _inferences = self.get_inferences(data)

            _out_readings = [{"asset": self.out_asset_name,
                              "readings": inference} for inference in _inferences]
            # now you have a list of new readings

            # send out a reading on the out asset
            if self.forward_data:
                # send out both the original and new inference readings
                data.extend(_out_readings)
            else:
                # only send out the list of new inference readings
                data = _out_readings

        # send out either our input data, our inferneces, or both
        filter_ingest.filter_ingest_callback(self.callback, self.ingest_ref, data)


###  PLUGIN INTERFACE  ###
_DEFAULT_CONFIG = {
    "plugin": {        # mandatory
        "description": "Filter plugin for ML inference on stream of image files",
        "type": "string",
        "default": "edgeml",
        "readonly": "true"
    },

    # ### filter-specific configuration ###
    "assetName": {
        "description": "Name of asset to listen on; readings have pathnames of images to analyze",
        "type": "string",
        "default": "edgeMl",
        "displayName": "Name of entries asset",
        "order": "1"
    },
    "outAssetName": {
        "description": "Name of asset to write ML inferences on",
        "type": "string",
        "default": "edgeMlInference",
        "displayName": "Name of inferences output asset",
        "order": "2"
    },
    "deploymentName": {
        "description": "Name of Kubernetes deployment for ML model",
        "type": "string",
        "default": "",
        "displayName": "Kubernetes deployment",
        "order": "3"
    },
    "edgeMlUrl": {
        "description": "REST URL for ML model which analyzes images; dynamically discovered if empty",
        "type": "string",
        "default": "",
        "displayName": "REST URL for ML model",
        "order": "4"
    },
    "forwardData": {
        "description": "Forward data as well as inferences",
        "type": "boolean",
        "default": "true",
        "displayName": "Forward inputs",
        "order": "5"
    },
    "rmFile": {
        "description": "Remove source files after inference",
        "type": "boolean",
        "default": "false",
        "displayName": "Remove files after inference",
        "order": "6"
    },
    "enable": {
        "description": "enable/disable plugin",
        "type": "boolean",
        "default": "true",
        "displayName": "Enable",
        "order": "7"
    },
}



def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        'name': 'edgeml',
        'version': '2.0.1',
        'type': 'filter',
        'interface': '1.0.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    _LOGGER.debug("Edge ML plugin_init called")
    _handle = None
    try:
        # convert config into an internal handle object; return a unique json-able handle
        _obj = EdgeMLHandle(config, ingest_ref, callback)
        _handle = _handle_map.new_handle(_obj, config)
    except Exception as ex:
        _LOGGER.exception('Error in initializing {} plugin {}'.format(__name__, str(ex)))
        raise

    return _handle


def plugin_ingest(h, data):
    """ plugin_ingest -- forward data to our callback if we are within our period """
    _handle = _handle_map.get_handle(h)
    _handle.ingest(data)
        

def plugin_reconfigure(h, new_config):
    """ Reconfigures the plugin
    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _handle = _handle_map.get_handle(h)
    _LOGGER.info("Old config for  plugin {} \n new config {}".format(h, new_config))
    ingest_ref, callback = _handle.ingest_ref, _handle.callback
    plugin_shutdown(h)
    # plugin_init
    new_handle = plugin_init(new_config, ingest_ref, callback)
    return new_handle


def plugin_shutdown(h):
    """ Shutdowns the plugin doing required cleanup
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _handle_map.del_handle(h)
